/***********************************************************************************
 *                                                                                 *
 * Matthew C. Berntsen, Bucknell University 2005                                   *
 * Thesis Project for Honors in Computer Science                                   *
 * Spring 2005                                                                     *
 *                                                                                 *
 * http://mattberntsen.net?page=thesis                                             *
 *                                                                                 *
 ***********************************************************************************/

#include "friedman.h"

// Pre:  None.
// Post: All internal variables are initialized
// Runtime: O(1);
friedman::friedman(unsigned int inMaxFile){
    // Set to accepted value
    EnglishIC = 0.065;
    estKeyLen = 0.0;
    maxFileLen = inMaxFile;
}

// Pre:  None.
// Post: return EnglishIC
// Runtime: O(1)
float friedman::getEnglishIC(){
    return EnglishIC;
}

// Pre:  inIC is a valid number 0 <= inIC <= 1
// Post: If inIC is valid, return the absolute value of the difference between
//       it and EnglishIC, else return -1.0.
// Runtime: O(1)
float friedman::diffEnglishIC(float inIC){

    // Validate input.
    if(inIC < 0.0 || inIC > 1.0)
	return -1.0;
    
    inIC -= EnglishIC;
    if(inIC < 0)
	inIC *= -1.0;
    return inIC;
}

// Pre:  Internal vars have been initialized properly and inFileName is a
//       valid string.
// Post: Calculate the approximate keylength, and set estKeyLen equal to
//       it. If successful, return true, else return false.
// Runtime: O(N) where N is the length of the input file.
bool friedman::calcKeyLen(string inFileName){

    // Declare and initialize variables.
    unsigned long int i, count[27];
    double countF[27];
    char tempC;
    for(i = 0; i < 27; i++)
	count[i] = 0;

    // Verify and open file.
    if(inFileName.size() > maxFileLen)
	return false;

    ifstream inFile(inFileName.c_str());
    if(!inFile.is_open())
	return false;

    // Collect Info
    while(inFile >> tempC){
	if (isspace(tempC))
	    ; //Do Nothing.
	else if(!isalpha(tempC)){
	    inFile.close();
	    return false;
	}
	else{ 
	    tempC = toupper(tempC);
	    if(int(tempC) - int('A') < 0 || int(tempC) - int('A') > 25)
	    printf("\t%s%c\n", "Error: ", tempC);
	count[int(tempC) - int('A')]++;
	count[26]++;
	}
    }

    // Done w/ input file.
    inFile.close();

    // Move data to floats to save on cast time. Use doubles to avoid possible overflow.
    for(i = 0; i < 27; i++){
	if(count[i] > 1)
	    countF[i] = double(count[i]);
	else
	    countF[i] = 1.1; // avoid divide by zero for n*(n - 1)
    }

    // Compute sample IC.
    prevIC = 0.0;
    for(i = 0; i < 26; i++){ // Done this way (slower) to avoid overflows.
	prevIC = prevIC + ((countF[i]/countF[26]) * ((countF[i] - 1.0)/(countF[26] - 1.0)));	
    }

     // Compute estimated key length.
    estKeyLen = (0.027 * count[26]) / (((count[26] - 1) * prevIC) - (0.038 * count[26]) + 0.065);
    return true;
}

// Pre:  estKeyLen is a valid float.
// Post: Return estKeyLen
// Runtime: O(1)
float friedman::getEstKeyLen(){
    return estKeyLen;
}


// Pre:  prevIC is a valid float.
// Post: Return prevIC
// Runtime: O(1)
float friedman::getPrevIC(){
    return prevIC;
}

// Pre:  Input variables are valid.
// Post: returns the average IC for the file obtained by dividing it into inLen
//       groups.
// Runtime: O(N + inLen)
float friedman::getIC(string inFileName, unsigned int inLen){
    if (inLen > 250 || inFileName.size() > 250)
	return -1.0;

    ifstream inFile(inFileName.c_str());
    if(!inFile.is_open())
	return -1.0;

    // Declare and initialize variables.
    unsigned int i, j;
    char tempC;
    unsigned long int ** matrix = new unsigned long int * [inLen];
    float * tempIC = new float[inLen];
    float * tempCount, rtnIC(0.0);
    for(i = 0; i < inLen; i++){ // O(inLen)
	matrix[i] = new unsigned long int [27];
	tempIC[i] = 0.0;
	for(j = 0; j < 27; j++)
	    matrix[i][j] = 0;
    }

    // Process frequency information.
    i = 0;
    while(inFile >> tempC){ // O(N)
	if(isspace(tempC))
	    ; // Do Nothing.
	else if(!isupper(tempC))
	    return -2.0;
	else{
	    matrix[i][int(tempC) - int('A')] ++;
	    matrix[i][26] ++;
	    i++;
	    i = i % inLen;
	}
    }

    // Calculate ICs.
    for(i = 0; i < inLen; i++){ // O(inLen)
	tempCount[i] = float(matrix[i][26]); // Cast in advance to save time.
	for(j = 0; j < 26; j++){ // O(1)
	    tempIC[i] = tempIC[i] + ((float(matrix[i][j])/float(tempCount[i])) * 
				     ((float(matrix[i][j]) + 1.0)/float((tempCount[i]) + 1.0)));
	}
	rtnIC += tempIC[i];
    }
    rtnIC = rtnIC / float(inLen);
    

    // Deallocate used memory.
    delete tempIC;
    for(i = 0; i < inLen; i++) // O(inLen)
	delete matrix[i];
    delete matrix;
    delete tempCount;

    // Return.
    return rtnIC;
}
