/***********************************************************************************
 *                                                                                 *
 * Matthew C. Berntsen, Bucknell University 2005                                   *
 * Thesis Project for Honors in Computer Science                                   *
 * Spring 2005                                                                     *
 *                                                                                 *
 * http://mattberntsen.net?page=thesis                                             *
 *                                                                                 *
 ***********************************************************************************/

#include "kasiski.h"
#include <cassert>

// Pre:  inputs are valid unsigned ints.
// Post: Internal variables are initialized.
// Runtime: O(1) as the number of primes is capped.
kasiski::kasiski(unsigned int inMaxFileLen, unsigned int inMaxPassLen){
    primes = NULL;
    weights = NULL;
    size3ch = 0xfffd; //0xfffd > 25^3 = 17576
    str3 = NULL;
    if(inMaxPassLen > 500 || inMaxPassLen < 2){
	printf("%s %d %s\n", "Error in kasiski.cc. inMaxPassLen is set to", 
	       inMaxPassLen, 
	       "however kaskski can handly only 2 <= inMaxPassLen <= 500.");
	exit(-1);
    }
    maxFileLen = inMaxFileLen;
    maxPassLen = inMaxPassLen;
    unsigned short tempS[] = {2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 
			      47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 101, 
			      103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 
			      157, 163, 167, 173, 179, 181, 191, 193, 197, 199, 
			      211, 223, 227, 229, 233, 239, 241, 251, 257, 263, 
			      269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 
			      331, 337, 347, 349, 353, 359, 367, 373, 379, 383, 
			      389, 397, 401, 409, 419, 421, 431, 433, 439, 443, 
			      449, 457, 461, 463, 467, 479, 487, 491, 499, 503};

    // ASSERT: maxPassLen <= 500
    for(numPrimes = 1; tempS[numPrimes] < maxPassLen; numPrimes++); // Find how many primes
    primes = new unsigned short[numPrimes];
    weights = new unsigned long int[numPrimes];

    // Move into propper data structures.
    for(unsigned short i(0); i < numPrimes; i++){
	primes[i] = tempS[i];
	weights[i] = 0;
    }	
}

// Pre:  All internal variables are valid.
// Post: All allocated memory is deallocated.
// Runtime: O(1) as the size of primes is capped.
kasiski::~kasiski(){
    if(primes!= NULL){
	delete primes;
	delete weights;
    }

    if(str3!= NULL){
	delete str3;
	for(unsigned int i = 0; i < size3ch; i++)
	    delIntN(loc3[i]);
	delete loc3;
    }
}

// Pre:  inFriedLen is a valid length, and internal variables are legal.
// Post: Determines (returns) most likely keylength based on inFriedLen and
//       primes/weights. If there is an error, returns USHRT_MAX.
// Runtime: O(1)
unsigned short kasiski::getResults(float inFriedLen){    
    
    // Declare variables.
    unsigned short min, max, inF = int(inFriedLen), i, j, tmpPrimes[numPrimes], tempS, end;
    unsigned long tmpWeights[numPrimes], tempL, mult;

    if(inFriedLen < 0.0){
	printf("Kasiski test requires that inFriedLen >= 0.0\n");
	return USHRT_MAX;
    }

    // Check to see that weights are not zeroed and copy into temp arrays
    for(i = 0; i < numPrimes; i++){
	if(weights[i] > 0)
	    break;
    }
    if(i == numPrimes) // Every weight was zero
	return USHRT_MAX;
    // If there is data to analyze, copy data into temp arrays to be moved around.
    for(i = 0; i < numPrimes; i++){ // O(1)
	tmpPrimes[i] = primes[i];
	tmpWeights[i] = weights[i];	
    }
    // Sort temp primes & weights lists
    for(i = 1; i < numPrimes; i++){ // O(1) as numPrimes is capped
	tempL = tmpWeights[i];
	tempS = tmpPrimes[i];
	for(j = i; j > 0 && tempL >= tmpWeights[j-1]; j--){ //O(1)
	    tmpWeights[j] = tmpWeights[j-1];
	    tmpPrimes[j] = tmpPrimes[j-1];
	}
	tmpWeights[j] = tempL;
	tmpPrimes[j] = tempS;
    }

    // Initialize variables.
    if (inF < int(inFriedLen + 0.5)) // Round up if applicable.
	inF++;
    if(inF < 25) // Set range to examine. It is <= 51 numbers.
	min = 1;
    else
	min = inF - 25;
    if(maxPassLen - 25 < inF)
	max = maxPassLen;
    else
	max = inF + 25;

    // Process
    // 2^9 = 512 > maxPassLen. Therefore anything weights less than 1/10 of
    // tmpWeights[0] is very unlikely to be in the pass length. 
    end = (tmpWeights[0] / 10) + 1; // Avoid zero bound.
    for(i = 1; i < numPrimes; i++)
	if(tmpWeights[i] < end)
	    break;
    end = i - 1; // end is now the last location to be considered.

//    printf("END: %d\n", end);
    assert(end <= numPrimes);
    tempS = 1;

    if(end == 0){ // Handle only one weights prime.
	while(tempS * tmpPrimes[0] < max)
	    tempS *= tmpPrimes[0];
	return tempS;
    }


    // Handle more than one weight.
    i = 0;
    j = i + 1;
    while(j <= end && i <= end && tmpWeights[j] > 0){ // O(1)
	assert(tmpWeights[i] >= tmpWeights[j]);	
	if(max/tmpPrimes[j] < tempS) // tmpPrimes[j] will go over max.
	    j++;
	if(tmpWeights[j] == 0)
	    break;
	    
	// Find multiplicative factor. If the remainder is >30% of the divisor,
	// round up.
	mult = tmpWeights[i] / tmpWeights[j];
	if(tmpWeights[i] % tmpWeights[j] > tmpWeights[j] / 3 && max / (mult + 1) > tempS)
	    mult++;
	if(max / mult > tempS)
	    tempS *= mult;
	if(max / tmpPrimes[j] > tempS){
	    tempS *= tmpPrimes[j];
	}

	// Advance placeholders.
	if(mult > 1){
	    i = j;
	    j = i + 1;
	}else
	    j++;
    }
    assert(tempS > 0 && tempS < max);   // Verify bounds. Multiply by a constant to bring as
    tempS *= (max/tempS);  // close to max as possible.

    // Return
    return tempS;
}

// Pre:  inFileName is a valid string
// Post: if file exists and contains at least 3 characters and has size > 500
//       characters, process and return true. Else return false.
// Runtime: O(N^2)
bool kasiski::process3(string inFileName){
    // Initialize variables to be used.
    ifstream inFile(inFileName.c_str());
    if(!inFile.is_open()){
	printf("File %s failed to open for read.\n", inFileName.c_str());
	return false;    
    }
    unsigned int i, j, k, listSize, loc(0), size1(0), size2(0), size3(0);
    intNode* locations[17576]; // 17576 = 26^3
    intNode* tempNode;
    for(i = 0; i < 17576; i++) // Loop O(1)
	locations[i] = NULL;
    for(i = 0; i < numPrimes; i++) // Loop O(1)
	weights[i] = 0;
    char in1, in2, in3;
    kasNode *begin, *cur, *beginOld, *curOld, *pos1, *pos2, *pos3;
    intNode* tempInt;
    pos1 = pos2 = pos3 = begin = cur = beginOld = curOld = NULL;
    tempInt = NULL;
    string tempS;
    
    // Print Dialog
    printf("%s%s\n", "Beginning Kaskski test on file: ", inFileName.c_str());
    
    // Find all values of length 3. Not used in computation, but might be
    // faster that length 4.
    if(!(inFile >> in1 >> in2 >> in3)){
	printf("Not Enough Characters in Input File to Kasiski Test\n");
	return false;
    }
    while(isspace(in1)){ // O(N)
	in1 = in2;
	in2 = in3;
	if(!inFile >> in3)
	    break;
	if(!isalpha(in3) && !isspace(in3)){
	    printf("Illegal characters in input file to Kaskski Test\n");
	    return false;
	}
    }
    while(isspace(in2)){ //O(N)
	in2 = in3;
	if(!inFile >> in3)
	    break;
	if(!isalpha(in3) && !isspace(in3)){
	    printf("Illegal characters in input file to Kaskski Test\n");
	    return false;
	}
    }
    if(!isalpha(in1) || !isalpha(in2) || !isalpha(in3)){
	printf("Illegal characters in input file to Kaskski Test\n");
	return false;
    }
    
    do{ // Loop O(N) where N is the size of the file.
	
	// Get rid of all spaces in input.
	while(isspace(in3)){ //O(N)
	    if(!inFile >> in3)
		break;
	    if(!isalpha(in3) && !isspace(in3)){
		printf("Illegal characters in input file to Kaskski Test\n");
		return false;
	    }
	}
	
	// Verify input validity
	if(isspace(in3))
	    break;
	if(!isalpha(in3)){
	    printf("Illegal characters in input file to Kaskski Test\n");
	    return false;
	}
	
	// Process
	i = toNum(in1, in2, in3);
	if(locations[i] == NULL)
	    locations[i] = new intNode(loc);
	else{
	    tempNode = new intNode(loc);
	    tempNode->next = locations[i];
	    locations[i] = tempNode;
	}
	
	in1 = in2;
	in2 = in3;
	loc++;
	if(loc == UINT_MAX - 1) // avoid overflow
	    break;
    }while(inFile >> in3);
    
    if(loc < 500){ // Make sure that at least 500 characters were analyzed.
	printf("%s %d %s\n", "The Kasiski test requires at least 500 characters of input. Only",
	       loc, "characters were analyzed.");
	return false;
    }
    
    //Find the number of repeated elements of length 3
    for(listSize = loc = 0; loc < 17576; loc++) // Loop O(1)
	if(locations[loc] != NULL && locations[loc]->next != NULL)
	    listSize++; // Cannot overflow, as capped at 17576
    
    if(listSize < 1){
	printf("No repeated elements to analize.\n");
	return false;
    }
    
    // Now copy those elements into a set of arrays. It is already in
    // lexicographical order.
    loc3 = new intNode*[listSize];
    str3 = new string[listSize];
    for(loc = j = 0; loc < 17576; loc++){ // O(1)
	if(locations[loc] != NULL && locations[loc]->next != NULL){
	    loc3[j] = locations[loc];
	    str3[j] = "---";
	    str3[j][0] = char(loc/676 + 'A');
	    str3[j][1] = char((loc/26)%26 + 'A');
	    str3[j][2] = char(loc % 26 + 'A');
	    j++;
	}
    }

    size3ch = listSize;
    assert(j == listSize && size3ch == listSize);

    // Print Dialog
    printf("%s %d %s\n", "There were", size3ch, "repeated substrings of length 3.");

    // Process Length 4
    listSize = 0;
    cur = begin = NULL;

    for(j = 0; j < size3ch; j++){ // This entire loop is O(N) as size3ch is capped.
	assert(str3[j].size() == 3);
	tempS = str3[j].substr(1);
	

	for(k = int('A'); k <= int('Z'); k++){ // O(1)
	    tempInt = compare(loc3[j], findStr(tempS + char(k)), 1); // O(N)
	    if(tempInt != NULL){ // O(1)
		listSize++;
		if(begin == NULL){
		    begin = cur = new kasNode(str3[j] + char(k), tempInt);
		}else{
		    cur->next = new kasNode(str3[j] + char(k), tempInt);
		    cur = cur->next;
		}
	    }
	}
    }

    // Create lists of subsequent sizes and compute weights
    for(i = 5; i <= maxPassLen && listSize > 0; i++){ // O(N^2)

	printf("%s %d %s %d%s\n", "There were", listSize, "repeated substrings of length", 
	       i-1, ".");
	listSize = 0;
	curOld = beginOld = begin;
	cur = begin = NULL;
	while(curOld != NULL){ // O(N^2)
	    tempS = curOld->data.substr(i - 3, 2);
	    for(k = 'A'; k <= 'Z'; k++){ //O(N)		
		tempInt = compare(curOld->locations, // O(N)
				  findStr(tempS + char(k)), i-3);
		if(tempInt != NULL){
		    if(listSize != UINT_MAX - 1)
			listSize++;
		    if(begin == NULL){
			begin = cur = new kasNode(curOld->data + char(k), tempInt);
		    }else{
			cur->next = new kasNode(curOld->data + char(k), tempInt);
			    cur = cur->next;
		    }
		}
	    }
	    curOld = curOld->next;
	}	    

	// Keep the last three substring lists in which subsequent ones are not
	// the same size to process weights later.
	if(listSize != size3){
	    delKas(pos1); // O(N^2)
	    pos1 = pos2;
	    pos2 = pos3;
	    pos3 = begin; // Most Recent, hence largest substring length.

	    size1 = size2;
	    size2 = size3;
	    size3 = listSize;
	}
    } 

    //Process Weights
    if(addWeights(pos1, 4) && addWeights(pos2, 2) &&  addWeights(pos3, 1))
	; // Do nothing. Verifys that it only runs until one of them overloads weights.

    // Deallocate used memory and return
    delKas(pos1); // Each is (N^2)
    delKas(pos2);
    delKas(pos3);
    for(i = 0; i < 17576; i++) // Whole process is O(N) where N is size of file.
	delIntN(locations[i]);

    return true;
}

// Pre:  inDel is a valid kasNode*
// Post: The linked list is deallocated.
// Runtime: O(N^2), where N^2 is the number of intNode*'s contained in the list.
void kasiski::delKas(kasNode* inDel){
    if(inDel == NULL)
	return;
    kasNode* del = inDel;
    kasNode* temp = inDel->next;
    while(temp != NULL){
	delIntN(del->locations);
	delete del;
	del = temp;
	temp = del->next;
    }
    delete del;	
    del = NULL;
}

// Pre:  inS is a string of size = 3
// Post: returns a numerical representation of the string.
// Runtime: O(1)
int kasiski::toNum(char inS0, char inS1, char inS2){
    if(!isalpha(inS0) || !isalpha(inS1) || !isalpha(inS2))
	return 0xffff;
    inS0 = toupper(inS0);
    inS1 = toupper(inS1);
    inS2 = toupper(inS2);

    unsigned int rtn = (int(inS0) - int('A')) * 676;
    rtn += (int(inS1) - int('A')) * 26;
    rtn += int(inS2) -  int('A');
    return rtn;
}

// Pre:  inS is a valid string
// Post: if str3 has been built, search it for the appropriate value. Return
// the location of the corresponding intNode* from loc3 if found, or return
// NULL if not found or error.
// Runtime: O(log(size3ch))
intNode* kasiski::findStr(string inS){
    if(inS.size() != 3 || size3ch == 0xFFFFFFFF)
	return NULL;
    unsigned int min(0), max(size3ch - 1), cur(size3ch / 2);
    while(min <= max){
	if(str3[cur] == inS) // If found, return.
	    return loc3[cur];
	if(inS < str3[cur]){ // Not found, continue w/ binary search.
	    if(cur == 0)
		return false;
	    max = cur - 1;
	}
	
	else
	    min = cur + 1;
	cur = (min + max) / 2;
    }
    return NULL;
}

// Pre:  inDel is a valid intNode*
// Post: The linked list is deallocated.
// Runtime: O(N)
void kasiski::delIntN(intNode * inN){
    if(inN == NULL)
	return;
    intNode* del = inN;
    intNode* temp = inN->next;
    while(temp != NULL){
	delete del;
	del = temp;
	temp = del->next;
    }
    delete del;
    del = NULL;
}

// Pre:  Input are valid variables
// Post: Traverses the two lists, and returns numbers that match with inDiff as
//       an offset if there are 2 or more in that list.
// Runtime: O(N) where N is the length of the shorter input linked list.
intNode* kasiski::compare(intNode* in1, intNode* in2, unsigned int inDiff){
    intNode* rtn, *cur, *temp1, *temp2;
    rtn = cur = NULL;
    temp1 = in1;
    temp2 = in2;
    while(temp1 != NULL && temp2 != NULL){ // O(N)
	if(temp1->data + inDiff == temp2->data){
	    if(rtn == NULL){
		rtn = new intNode(temp1->data);
		cur = rtn;
	    }else{
		cur->next = new intNode(temp1->data);
		cur = cur->next;
	    }
	    temp1 = temp1 -> next;
	    temp2 = temp2 -> next;
	}else if(temp1->data + inDiff < temp2->data)
	    temp2 = temp2->next;
	else
	    temp1 = temp1->next;
    }

    if(rtn == cur) // If < 2 locations in list
	return NULL;
    return rtn;
}

// Pre:  inBegin and mass are valid variables & weights is initialized to zero.
// Post: Runs through the matrices of locations are computes the prime numbers
//       that are the most common, adding mass every time it finds
//       one. Returns true if no weights are overflowed, else returns false.
bool kasiski::addWeights(kasNode* inBegin, unsigned short mass){
    intNode *refInt, *curInt;
    unsigned long int len, tempI;
    unsigned short tempS;
    for(kasNode* cur = inBegin; cur != NULL; cur = cur->next){
	for(refInt = curInt = cur->locations; refInt!= NULL && refInt->next != NULL;
	    refInt = refInt->next){ //O(N^2)
	    for(;curInt != NULL; curInt = curInt ->next){ //O(N)
		tempI = refInt->data - curInt->data;
		for(tempS = 0; tempS < numPrimes; tempS++){  //O(1)
		    len = tempI;
		    while(len > 0 && len % primes[tempS] == 0){
			if(weights[tempS] >= ULONG_MAX - mass) // Check for overflow.
			    return false;
			weights[tempS] += mass;
			len = len / primes[tempS];
		    }
		}
	    }
	}
    }
    return true;
}

      
    
