/***********************************************************************************
 *                                                                                 *
 * Matthew C. Berntsen, Bucknell University 2005                                   *
 * Thesis Project for Honors in Computer Science                                   *
 * Spring 2005                                                                     *
 *                                                                                 *
 * http://mattberntsen.net?page=thesis                                             *
 *                                                                                 *
 ***********************************************************************************

 Description: This class performs all funtions of the Kasiski test, as well as
              combining its results with the result of the Friedman test to
              determine the best-fit keylength.

 ***********************************************************************************/

#ifndef __KASISKI_H
#define __KASISKI_H

using namespace std;
#include <string>
#include <fstream>

struct intNode{
  unsigned int data;
  intNode * next;
  
  // Pre:  None.
  // Post: Initialize internal variables
  // Runtime: O(1)
  intNode(){
    next = NULL;
  }

  // Pre:  None.
  // Post: Initialize internal variables
  // Runtime: O(1)
  intNode(int inData){
    next = NULL;
    data = inData;
  }
};

struct kasNode{ // Kasiski Node. Stores a string w/ a list of locations.
  string data;
  intNode* locations;
  kasNode* next;

  // Pre:  None.
  // Post: Initialize internal variables
  // Runtime: O(1)
  kasNode(){
    locations = NULL;
    next = NULL;
  }

  // Pre:  None.
  // Post: Initialize internal variables
  // Runtime: O(1)
  kasNode(string inData){
    data = inData;
    locations = NULL;
    next = NULL;
  }

  // Pre:  None.
  // Post: Initialize internal variables
  // Runtime: O(1)
  kasNode(string inData, intNode* inLoc){
    data = inData;
    locations = inLoc;
    next = NULL;
  }
};

class kasiski{

public:

    // Pre:  inputs are valid unsigned ints.
    // Post: Internal variables are initialized.
    // Runtime: O(1) as the number of primes is capped.
    kasiski(unsigned int inMaxFileLen, unsigned int inMaxPassLen);

    // Pre:  All internal variables are valid.
    // Post: All allocated memory is deallocated.
    // Runtime: O(1) as the size of primes is capped.
    ~kasiski();

    // Pre:  inFileName is a valid string
    // Post: if file exists and contains at least 3 characters and has size > 500
    //       characters, process and return true. Else return false.
    // Runtime: O(N^2)
    bool process3(string inFileName);
 
    // Pre:  inFriedLen is a valid length, and internal variables are legal.
    // Post: Determines (returns) most likely keylength based on inFriedLen and
    //       primes/weights. If there is an error, returns USHRT_MAX.
    // Runtime: O(1)
    unsigned short getResults(float inFreidLen);

private:
    // PRIVATE FUNCTIONS

    // Pre:  inDel is a valid kasNode*
    // Post: The linked list is deallocated.
    // Runtime: O(N^2), where N is the number of intNode*'s contained in the list.
    void delKas(kasNode* inDel);

    // Pre:  inS is a string of size = 3
    // Post: returns a numerical representation of the string.
    // Runtime: O(1)
    int toNum(char inS0, char inS1, char inS2);

    // Pre:  inDel is a valid intNode*
    // Post: The linked list is deallocated.
    // Runtime: O(N)
    void delIntN(intNode * inN);

    // Pre:  inS is a valid string
    // Post: if str3 has been built, search it for the appropriate value. Return
    // the location of the corresponding intNode* from loc3 if found, or return
    // NULL if not found or error.
    // Runtime: O(log(size3ch))
    intNode* findStr(string inS);

    // Pre:  Input are valid variables
    // Post: Traverses the two lists, and returns numbers that match with inDiff as
    //       an offset if there are 2 or more in that list.
    // Runtime: O(N) where N is the length of the shorter input linked list.
    intNode* compare(intNode* in1, intNode* in2, unsigned int inDiff);

    // Pre:  inBegin and mass are valid variables & weights is initialized to zero.
    // Post: Runs through the matrices of locations are computes the prime numbers
    //       that are the most common, adding mass every time it finds
    //       one. Returns true if no weights are overflowed, else returns false.
    // Runtime: O(N^2)
    bool addWeights(kasNode* inBegin, unsigned short mass);\

    // PRIVATE VARIABLES
    unsigned short * primes;     // List of prime numbers in ascending order

    unsigned long int * weights; // Corresponding weights to above primes

    unsigned int numPrimes;      // The number of primes in the list above.

    unsigned int maxFileLen;     // Max length of filename to be opened

    unsigned int maxPassLen;     // Max keylength to be examined

    intNode** loc3;              // Matrix of locations corresponding to every 
                                 // repeated 3-character substring.

    string* str3;                // 3-letter substrings that corresponed to loc3

    unsigned int size3ch;          // Size of loc3
};

#endif
