/***********************************************************************************
 *                                                                                 *
 * Matthew C. Berntsen, Bucknell University 2005                                   *
 * Thesis Project for Honors in Computer Science                                   *
 * Spring 2005                                                                     *
 *                                                                                 *
 * http://mattberntsen.net?page=thesis                                             *
 *                                                                                 *
 ***********************************************************************************/

#include "stringList.h"

// Pre:  None.
// Post: Set initial values.
// Runtime: O(1)
stringList::stringList(){
    numElements = 0;
    start = end = NULL;
    prepDone = false;
    dataStr = NULL;
}


// Pre:  List contents are contiguous and valid. if(loc), for all list elements
//       pos != NULL.
// Post: All allocated memory is deallocated.
// Runtime: O(numElements)
stringList::~stringList(){
    // Assert, list contents are contiguous.
    if(numElements > 0){
	if(prepDone)
	    delete dataStr;
	else{
	    strListNode* temp,* tempDel;
	    tempDel = temp = start;
	    while(temp != NULL){
		temp = temp->next;
		delete tempDel;
		tempDel = temp;
	    }
	}
    }
}

// Pre:  numElements is a valid unsigned int.
// Post: return numElements
// Runtime: O(1)
unsigned int stringList::size(){
    return numElements;
}

// Pre:  input is a valid string.
// Post: inS is inserted into the list.
// Runtime: O(1) as it is a linked list insertion.
bool stringList::insert(string inS){
    // Wrong state.
    if(numElements == 0xfffffffe || prepDone)
	return false;

    // Caste to upper.
    for(unsigned int i = 0; i < inS.size(); i++)
	inS[i] = toupper(inS[i]);
    
    if(start == NULL){ // First element
	start = end = new strListNode;
    }

    else if(end->data < inS){
	end->next = new strListNode;
	end = end->next;
    }
    // Not an end insertion. Bail.
    else{
	printf("%s %s %s\n", "Insertion of \"", inS.c_str(), 
	       "\" failed as it is out of order.");
	return false;
    }

    end->data = inS;
    end->next = NULL;
    numElements++;
    return true;
}

// Pre:  inS is a valid string and the list is legally constructed.
// Post: searches the list for inS. Returns true if it finds it, else false.
// Runtime: O(log(numElements))
bool stringList::search(string inS){
    return (find(inS) != 0xffffffff);
}

// Pre:  inS is a valid string and the list is legally constructed.
// Post: searches the list for inS. Returns the index if it finds it, else 0xffffffff
// Runtime: O(N) the first time it is run (runs prep()), and O(Log(N)) thereafter.
unsigned int stringList::find(string inS){
    for(unsigned int i = 0; i < inS.size(); i++)
	inS[i] = toupper(inS[i]);
    
    if(!prepDone)
	prep();

    unsigned int tempLoc = numElements/2, tempMax = numElements - 1, tempMin = 0;
    
    while(tempMin <= tempMax){
	// ASSERT: tempMin <= tempLoc <= tempMax
	// Found, return.
	if(inS == dataStr[tempLoc])
	    return tempLoc;
	
	// In lower segment.
	if(inS < dataStr[tempLoc]){
	    tempMax = tempLoc - 1;
	}else{ //inS > data[tempLoc].data
	    tempMin = tempLoc + 1;
	}
	tempLoc = (tempMin + tempMax) / 2;	    
	
	if(inS == dataStr[tempLoc])
	    return tempLoc;  
    }
    return 0xffffffff;
}

// Pre:  The list is configured properly.
// Post: Prints the list string contents.
// Runtime: O(numElements)
void stringList::print(){
    if(prepDone)
	for(unsigned int i = 0; i < numElements; i++)
	    printf("%s\n", dataStr[i].c_str());
    else{
	strListNode* temp = start;
	while(temp != NULL){
	    printf("%s\n", temp->data.c_str());
	    temp = temp->next;
	}
    }
}

// Pre:  All internal data is valid.
// Post: The linked list of data is transfered into an array for searching. 
//       prepDone = true and start & end are deleted and set to NULL.
// Runtime: O(numElements)
void stringList::prep(){
    if(!prepDone){ // Verify that it can be run
	dataStr = new string[numElements];
	strListNode* temp,* tempDel;
	temp = tempDel = start;

	// ASSERT: numElements reflect # of elements in list starting at start
	for(unsigned int i = 0; i < numElements && temp != NULL; i++){
	    dataStr[i] = temp->data;
	    temp = temp->next;
	    delete tempDel;
	    tempDel = temp;
	}
	start = end = NULL;
	prepDone = true;
    }
}
