/***********************************************************************************
 *                                                                                 *
 * Matthew C. Berntsen, Bucknell University 2005                                   *
 * Thesis Project for Honors in Computer Science                                   *
 * Spring 2005                                                                     *
 *                                                                                 *
 * http://mattberntsen.net?page=thesis                                             *
 *                                                                                 *
 ***********************************************************************************

 Description: This file is the driver for the project. Its usage is as follows:
 
     Usage: vigenere [ACTION] [MODE] [PASSPHRASE] [FILENAME]

     ACTION is either '-e' for Encode, '-d' for Decode, '-c' for Crack,
             or -p for Prep. For -c or -p, the passphrase is omitted.

     MODE is an integer, starting at 1, that defines which method of
             operation will be used for the given function. For -p
             MODE is 1 (include whitespace) or 2 (exclude whitespace).
             For -e and -d, mode 1 uses the tableau method and mode 2
             uses the computational method of the vigenere cipher. For
             -c, mode 1 is the Friedman test only, mode 2 uses the
             Kasiski (and Friedman) test, and mode 3 is the results from
             both.

     PASSPHRASE is a series of letters used to encode or decode
             the specified file. 

     FILENAME is the full or relative path and filename of the ASCII
             text file to be operated upon.

 ***********************************************************************************/

// Includes
using namespace std;
#include <string>
#include <fstream>
#include "vigCrypt.h"
#include "friedman.h"
#include "kasiski.h"
#include "crackMShift.h"
#include "dictLookup.h"
#include <cassert>

// Function Declarations
void argError(string, string);
void argError(string inProg){argError(inProg, "");}
bool vigPrep(char*, bool);

// Define variables used to cap string lengths to prevent buffer overflows.
const int MAX_PASS_LENGTH = 200;
const int MAX_FILENAME_LENGTH = 150;

// Runtime: O(N^2)
// Where N is the number of characters in the input file
// This assumes that N is the size of the input file and m is the size of the
// passphrase if applicable.
int main(int argc, char* argv[]){
    // DECLARE VARIABLES
    int temp;
    char tempC;

    // VERIFY COMMANDLINE PARAMATERS [O(1)]
    if(argc < 3 || argc > 5 || strlen(argv[1]) != 2 || argv[1][0] != '-' || 
       ((tempC = tolower(argv[1][1])) != 'e' && tempC != 'd' && tempC != 'c' 
	&& tempC != 'p'))
	argError(argv[0], "Invalid Commandline Parameters");
    
    // CRACK
    if(argv[1][1] == 'c' && argc == 4 && strlen(argv[2]) == 1 && argv[2][0] > '0' && argv[2][0] < '4'){
	
	// Declare and initialize variables
	friedman theFried(MAX_FILENAME_LENGTH);
	kasiski theKas(MAX_FILENAME_LENGTH, MAX_PASS_LENGTH);
	vigCrypt theCrypt(MAX_FILENAME_LENGTH, MAX_PASS_LENGTH);
	crackMShift theCrack(MAX_PASS_LENGTH, MAX_FILENAME_LENGTH);
	dictLookup theDict;
	float friedResult, tempF, tempF2;
	unsigned int kasI, mode(atoi(argv[2])), friedResI, min, max, i;
	string friedKey, kasKey, dec(argv[3]), tempS;
	dec = dec + ".dec";

	// GET DATA
	if(!theFried.calcKeyLen(argv[3])){
	    printf("%s%s\n", "Friedman test unable to run on file: ", argv[3]);
	    exit(-1);
	}
	
	friedResult = theFried.getEstKeyLen();
	if(mode != 2){
	    // Perform Friedman Test Analysis.
	    friedResI = (unsigned int)(friedResult);
	    
	    // Round.
	    if(friedResI < (unsigned int)(friedResult + 0.5))
		friedResI++;

	    // Handle boundaries.
	    if(friedResI < 28)
		min = 2;
	    else
		min = friedResI - 25;
	    
	    if(friedResI > MAX_PASS_LENGTH - 25){
		max = MAX_PASS_LENGTH;
		min = MAX_PASS_LENGTH - 30;
	    }else
		max = friedResI + 25;
	    
	    assert(min < max);

	    tempF = theFried.getIC(argv[3], min);
	    friedResI = min;

	    for(i = min + 1; i <= max; i++){
		tempF2 = theFried.getIC(argv[3], i);
		if(theFried.diffEnglishIC(tempF) > theFried.diffEnglishIC(tempF2)){
		    friedResI = i;
		    tempF = tempF2;
		}
	    }

	    if(theCrack.crack(friedResI, argv[3]))
	    friedKey = theCrack.getPass();
	}

	if(mode != 1){
	    if(!theKas.process3(argv[3])){
		printf("%s%s\n", "\tKasiski test unable to run on file: ", argv[3]);
		exit(-1);	
	    }
	    kasI = theKas.getResults(friedResult);
	    if((mode == 2 || friedResI != kasI) && theCrack.crack(kasI, argv[3]))
	    kasKey = theCrack.getPass();
	}
	printf("\nRESULTS:\n");
	// PROCESS DATA
	switch(mode){
	case 1:
	    printf("\n\tFriedman test best-fit keylength: %d\n", friedResI);
	    printf("\tBest-fit key: %s\n", friedKey.c_str());
	    theCrypt.crypt(2, false, friedKey, argv[3]);
	    tempF = theDict.percentEnglish(dec);
	    if(tempF >= 0.0)
		printf("\tUsing the key %s, the output is %f%sEnglish words.\n", friedKey.c_str(), tempF, "% ");
	    break;

	case 2:
	    printf("\n\tKasiski test best-fit keylength: %d\n", kasI);
	    printf("\tBest-fit key: %s\n", kasKey.c_str());
	    theCrypt.crypt(2, false, kasKey, argv[3]);
	    tempF = theDict.percentEnglish(dec);
	    if(tempF >= 0.0)
	    printf("\tUsing the key %s, the output is %f%sEnglish words.\n", kasKey.c_str(), tempF, "% ");
	    break;

	case 3:
	    tempS = "";
	    // Friedman key will always be <= kasKey.
	    assert(kasKey.size() > friedKey.size());
	    if(friedKey.size() < kasKey.size() && kasKey.size() % friedKey.size() == 0){
		max = kasKey.size() / friedKey.size();
		for(i = 0; i < max; i++)
		    tempS = tempS + friedKey;
		if(kasKey == tempS)
		    kasKey = friedKey;
	    }
		    
	    if(kasKey.size() == friedKey.size()){
		printf("\n\tBoth tests suggest a keylength of: %d\n", friedKey.size());
		printf("\tBest-fit key: %s\n", friedKey.c_str());
		theCrypt.crypt(2, false, friedKey, argv[3]);
		tempF = theDict.percentEnglish(dec);
		if(tempF >= 0.0)
		    printf("\tUsing the key %s, the output is %f%sEnglish words.\n", friedKey.c_str(), tempF, "% ");
	    }else{
		printf("\n\n\tFriedman test best-fit keylength: %d\n", friedResI);
		printf("\tFriedman test best-fit key: %s\n", friedKey.c_str());
		theCrypt.crypt(2, false, friedKey, argv[3]);
		tempF = theDict.percentEnglish(dec);
		if(tempF >= 0.0)
		    printf("\tUsing the key %s, the output is %f%sEnglish words.\n", friedKey.c_str(), tempF, "% ");
		printf("\tKasiski test best-fit keylength: %d\n", kasI);
		printf("\tKasiski test best-fit key: %s\n", kasKey.c_str());
		theCrypt.crypt(2, false, kasKey, argv[3]);
		tempF = theDict.percentEnglish(dec);
		if(tempF >= 0.0)
		    printf("\tUsing the key %s, the output is %f%sEnglish words.\n", kasKey.c_str(), tempF, "% ");
	    }
	    break;
	}

	printf("\n");
	// Remove created file.
	dec = "rm " + dec;
	system(dec.c_str());


	/**************************** FUNCTIONALITY TESTS **********************
          This section below included for testing purposes if this code is ever
          taken up and modified. These ignore the value for mode.
        ************************************************************************/

	/* TEST KASISKI w/ FRIEDMAN TESTS 
	kasiski theKas(MAX_FILENAME_LENGTH, MAX_PASS_LENGTH);
	friedman theFried(MAX_FILENAME_LENGTH);
	
	if(theKas.process3(argv[3]) && theFried.calcKeyLen(argv[3])){
	    float tmpFried = theFried.getEstKeyLen();
	    printf("TheKas Successful, Best Fit: ");
	    unsigned short tmpS = theKas.getResults(tmpFried);
	    printf("%d\nFried: %f\n", tmpS, tmpFried);
	}
	else
	    printf("TheKas Failed.\n");
	*/

 	/* DICTIONARY LOOKUP TEST *
 	ifstream inFile(argv[3]);
 	if(inFile.is_open()){
 	    string inWord;
 	    unsigned int count(0);
 	    dictLookup theD;
 	    while(inFile >> inWord){
 		theD.lookup(inWord);
 		count ++;
 	    }
 	    printf("%s %d\n", "Total Words:", count);
 	    printf("%s %d\n", "Dict Size:", theD.size());
 	}
	*/

 	/* MULTIPLE SHIFT TEST CRACK *
 	crackMShift theCMS(MAX_PASS_LENGTH, MAX_FILENAME_LENGTH);
 	theCMS.crack(5 , argv[3]);
 	printf("\t%s%s\n", "Best Fit Passphrase (5ch): ", theCMS.getPass().c_str());
 	*/

 	/* FRIEDMAN TEST TEST *
 	friedman theFried(MAX_FILENAME_LENGTH);
 	theFried.calcKeyLen(argv[3]);
 	printf("\t%s %f \n", "PrevIC:", theFried.getPrevIC());
 	*/

 	/* SINGLE SHIFT CRACK TEST *
 	crackShift theCS;
 	theCS.crack(argv[3]);
 	printf("%s%d\n", "Best-Fit Shift: ", theCS.getShift());
 	*/

 	/***********************************************************************/
     }


    // ENCRYPT/DECRYPT [O(N)]
    else if((tolower(argv[1][1]) == 'e' || tolower(argv[1][1]) == 'd') && 
 	    argc == 5 && (temp = strlen(argv[2])) < MAX_PASS_LENGTH){
 	int inMode = atoi(argv[2]);
 	if (inMode < 1 || inMode > 2)
 	    argError(argv[0], "Invalid Commandline Parameters");  
 	// Validate passphrase argument
 	for(int i = 0; i < temp; i++)
 	    if(!isalpha(argv[3][i]))
 		argError(argv[0], "Illegal Characters in Passphrase.");
 	vigCrypt theCrypt(MAX_FILENAME_LENGTH, MAX_PASS_LENGTH);
	theCrypt.crypt(inMode, (argv[1][1] == 'e') /*Incrypt or Decrypt*/,
		       argv[3] /*Passphrase*/, argv[4] /*Filename*/); 
    }


    // PREP [O(N)]
    else if(argv[1][1] == 'p' && argc == 4){ 
	if(strlen(argv[2]) == 1 && (argv[2][0] == '1' || argv[2][0] == '2')){
	    vigPrep(argv[3], argv[2][0] == '1');
	}
	else
	    argError(argv[0], "Invalid Commandline Parameters");  
    }


    // FAIL [O(1)]
    else
	argError(argv[0], "Invalid Commandline Parameters");    
}

// Pre: inFileName is a valid char*
// Post: the appropriate file is opened if possible (it exists and access is
// allowed), and is read into inFileName.vigPrep dropping all non-letters and
// forcing all letter to upper case. if(ws), then all whitespace is resolved to
// a ' ' character. Returns true if successful, else false.
// Runtime: O(N) where N is the number of characters in the file being read.
bool vigPrep(char* inFileName, bool ws){
    if(strlen(inFileName) > 250) //Check for buffer overflow in filename
	printf("%s%s", "\nThe specified path is greater than 250 characters in length.\n",
	       "Try using a relative path.\n\n");
    ifstream inFile;
    ofstream outFile;
    char tempC;
    bool prevSpace(false);

    inFile.open(inFileName);
    if (!inFile.is_open()){ // Cannot read input file.
	printf("%s%s\n\n", "\nCannot read from file: ", inFileName);
	return false;
    }
    
    outFile.open((string(inFileName) + ".vigPrep").c_str(), ios::out);
    if (!outFile.is_open()){ // Cannot open output file for writing.
	printf("%s%s%s\n\n", "\nCannot open for writing: ", inFileName, ".vigPrep");
	return false;
    }

    // Write letters only in uppercase from inFile to outFile.
    while(inFile.read(&tempC, 1)){
	if (ws && isspace(tempC)){
	    if(!prevSpace){ //Do not allow multiple spaces.
		outFile.put(' ');
		prevSpace = true;
	    }
	}
	else if(isalpha(tempC)){
	    outFile.put(toupper(tempC));
	    prevSpace = false;
	}
    }

    inFile.close();
    outFile.close();
    return true;
}

// Pre: progName and message are valid char* data types.
// Post: Display program usage and message if applicable.
// Runtime: O(1)
void argError(string progName, string message){
    if (message.length() > 0)
	printf("\n\t%s\n\n", message.c_str());

    printf("%s%s%s%s%s%s%s%s%s%s%s%s%s%s\n\n", "Usage: ", progName.c_str(), 
	   " [ACTION] [MODE] [PASSPHRASE] [FILENAME]\n\n",
	   "ACTION is either '-e' for Encode, '-d' for Decode, '-c' for Crack,\n",
	   "\tor -p for Prep. For -c or -p, the passphrase is omitted.\n\n",
	    "MODE is an integer, starting at 1, that defines which method of\n",
	   "\toperation will be used for the given function. For -p\n",
	   "\tMODE is 1 (include whitespace) or 2 (exclude whitespace).\n",
	   "\tFor -e and -d, mode 1 uses the tableau method and mode 2\n",
	   "\tuses the computational method of the vigenere cipher. For\n",
	   "\t-c, mode 1 is the Friedman test only, mode 2 uses the\n"
	   "\tKasiski (and Friedman) test, and mode 3 is the results from\n\tboth.\n\n",
	   "PASSPHRASE is a series of letters used to encode or decode\n\tthe ",
	   "specified file. \n\nFILENAME is the full or relative path and filename of",
	   " the ASCII\n\ttext file to be operated upon.");
    exit(-1);
}
